from PyQt5 import QtSql
from enum import Enum
from typing import Any, Dict, Union, List, Tuple
from . import logger


default_databases = [{'title': 'SQLite3', 'driver': 'QSQLITE', 'name': 'athnos.sqlite3', 'user': None, 'password': None, 'ip': None, 'port': None},
                     {'title': 'MariaDB', 'driver': 'QMARIADB', 'name': 'athnos', 'user': 'root', 'password': '', 'ip': '127.0.0.1', 'port': 3306},
                     {'title': 'MySQL', 'driver': 'QMYSQL', 'name': 'athnos', 'user': 'root', 'password': '', 'ip': '127.0.0.1', 'port': 3306},
                     {'title': 'PostgreSQL', 'driver': 'QPSQL', 'name': 'athnos', 'user': 'postgres', 'password': '', 'ip': '127.0.0.1', 'port': 5432},
                     {'title': 'ODBC', 'driver': 'QODBC', 'name': 'athnos', 'user': 'postgres', 'password': '', 'ip': '127.0.0.1', 'port': 1433},
                     {'title': 'IBM DB2', 'driver': 'QDB2', 'name': 'athnos', 'user': 'db2admin', 'password': 'db2admin', 'ip': '127.0.0.1', 'port': 50000},
                     {'title': 'Borland InterBase', 'driver': 'QIBASE', 'name': 'athnos', 'user': '', 'password': '', 'ip': '127.0.0.1', 'port': 3050}]


class _Table:
    """Base class of every table in `Database`"""

    def __init__(self, database, table_name: str, zero_id_value=None):
        """
        Args:
            database (Database): Database of the table
            table_name: Name of the table
        """
        self.database = database
        self.table_name = table_name
        self.zero_id_value = zero_id_value

        # checks if the table exists
        query = self.new_query()
        query.exec(f'SELECT 1 FROM {self.table_name}')

        if query.lastError().isValid():
            # raises an IOError if the table doesn't exist
            raise IOError(query.lastError().text())
        else:
            # adds all table columns to a list
            self.columns = []
            column_record = self.database.record(table_name)
            for column in range(column_record.count()):
                # » reads the column names and adds it
                self.columns.append(column_record.field(column).name())

        # sets `self.item` and `self.rows`.
        # `self.item` is a child class of `_Row`, `self.rows` is a child class of `self._Rows`
        if issubclass(self.__class__, Database._Clip):
            self.item = Clip
            self.rows = ClipRows
        elif issubclass(self.__class__, Database._Item):
            self.item = Item
            self.rows = ItemRows
        elif issubclass(self.__class__, Database._Source):
            self.item = Source
            self.rows = SourceRows
        else:
            self.item = Tag
            self.rows = TagRows

    def delete(self, id: int) -> None:
        """
        Deletes the entry with the corresponding id from the database

        Args:
            id: ID of the entry to delete
        """
        query = self.new_query()
        query.exec(f'DELETE FROM {self.table_name} WHERE id={id}')
        if query.lastError().isValid():
            logger.warning(f'Error while deleting {id} from {self.table_name}: {query.lastError().text()}')

    def edit(self, id: int, **kwargs) -> None:
        """
        Edits the entry with the corresponding id from the database

        Args:
            id: ID of the entry to edit
            **kwargs: New values for specific columns

        Raises:
            ValueError: if one of the `**kwargs` is not in `self.columns` or is ``id``
        """
        self.make_sql_valid(kwargs)
        query = self.new_query()
        query.exec(f'UPDATE {self.table_name} SET {", ".join([f"{column}={value}" for column, value in kwargs.items()])} WHERE id={id}')
        if query.lastError().isValid():
            logger.warning(f'Error while updating {id} in {self.table_name}: {query.lastError().text()}')

    def get(self, id: int):
        """
        Get a database object by it's id

        Args:
            id: The id of the object you want to get

        Returns:
            [Clip, Item, Source, Tag]: The database object

        Raises:
            QtSql.QSqlError: If the sql query wasn't executed correct
        """
        if id == 0:
            # returns a special value if `id` is 0 and `self.zero_id_value` is not None
            if self.zero_id_value:
                return self.zero_id_value
        query = self.new_query()
        query.exec(f'SELECT * FROM {self.table_name} WHERE id={id}')
        self.validate_query(query, True)
        return self.item(self.database, query)

    def get_all(self):
        """
        Get all infos in the table

        Returns:
            _Rows: The infos stored in a child row from ``_Rows``
        """
        query = self.new_query()
        query.exec(f'SELECT * FROM {self.table_name}')
        self.validate_query(query, True)
        return self.rows(self.database, query=query)

    def get_all_by(self, **kwargs):
        """
        Get items by its values

        Args:
            **kwargs: The values from the items you want to get

        Returns:
            The information stored in a child class of ``_Rows``
        """
        self.make_sql_valid(kwargs)
        query = self.new_query()
        query.exec(f'SELECT * FROM {self.table_name} WHERE {" AND ".join(f"{column_name}={value}" for column_name, value in kwargs.items())}')
        self.validate_query(query, True)
        return self.rows(self.database, query=query)

    def get_by(self, **kwargs):
        """
        Get a item by its values

        Args:
            **kwargs: The values from the item you want to get

        Returns:
            The information stored in a child class of ``_Row``
        """
        self.make_sql_valid(kwargs)
        query = self.new_query()
        query.exec(f'SELECT * FROM {self.table_name} WHERE {" AND ".join(f"{column_name}={value}" for column_name, value in kwargs.items())}')
        self.validate_query(query, True)
        return self.item(self.database, query=query)

    def has(self, **kwargs) -> bool:
        """
        Checks if the table has a entry, corresponding with the `**kwargs`

        Args:
            include_missing: Includes all column names (besides `id`), which weren't specified in `**kwargs` and set their value to None / NULL
            **kwargs: Table keys and their values to check if they exists in the table

        Returns:
            If the `**kwargs` are in the table
        """
        self.make_sql_valid(kwargs)
        query = self.new_query()
        query.exec(f'SELECT 1 FROM {self.table_name} WHERE ({" AND ".join(f"{key}={value}" for key, value in kwargs.items())})')
        self.validate_query(query, True)
        if query.isValid():
            return True
        else:
            return False

    def has_id(self, id: int) -> bool:
        """
        Checks if the table has a given id

        Args:
            id: ID to check if its in the table

        Returns:
            If the id is in the table
        """
        return self.has(id=id)

    def highest_id(self) -> int:
        """
        Selects the highest / max id from the given `self.table_name`

        Returns:
            The highest id

        """
        query = self.new_query()
        query.exec(f'SELECT MAX(id) FROM {self.table_name}')
        query.next()

        # the highest id
        highest_id = query.value(0)

        if highest_id:
            return highest_id
        else:
            return 0

    def new_query(self) -> QtSql.QSqlQuery:
        """
        Get a new sql query

        Returns:
            The sql query

        """
        return QtSql.QSqlQuery(self.database)

    def make_sql_valid(self, to_correct: Union[Dict[Any, Any], List[Any]]) -> Union[Dict[Any, Any], List[Any]]:
        """
        Makes all items in a dictionary or list valid to use in a sql query

        Args:
            to_correct: The dict or list to make valid

        Returns:
            The corrected dict or list

        Raises:
            ValueError: If one of the items in not str, int or bool
        """
        if isinstance(to_correct, dict):
            for key, value in to_correct.items():
                if isinstance(value, str):
                    to_correct[key] = f'"{value}"'
                elif isinstance(value, int):
                    to_correct[key] = str(value)
                elif not isinstance(value, bool):
                    raise ValueError(f'The value {value} of key {key} must be type str, int or bool')
        elif isinstance(to_correct, list):
            for i, item in enumerate(to_correct):
                if isinstance(item, str):
                    to_correct[i] = f'"{item}"'
                elif isinstance(item, int):
                    to_correct[i] = str(item)
                elif not isinstance(item, bool):
                    raise ValueError(f'{item} must be type str, int or bool')

        return to_correct

    def validate_query(self, query: QtSql.QSqlQuery, error=False) -> bool:
        """
        Checks if the given query has errors and if it has, a log output will be generated

        Args:
            query: Query to check
            error: If given, instead of the log output, a error will be raised

        Returns:
            If the query has errors or not

        Raises:
            IOError: If the query has errors and `error` is True
        """
        if query.lastError().isValid():
            message = f'Error while executing query \'{query.lastQuery()}\': {query.lastError().text()}'
            if error:
                raise IOError(message)
            else:
                logger.warning(message)
            return False
        return True


class Type:
    """Enum for clip / source types"""
    class Clip(Enum):
        OTHER = 0
        AUDIO = 1
        IMAGE = 2
        VIDEO = 3

        @staticmethod
        def from_name(name: str):
            """
            Returns a `Type.Clip` by its name

            Args:
                name: Name of the value

            Returns:
                Type.Clip: The corresponding type
            """
            for type in Type.Clip:
                if type.name.lower() == name.lower():
                    return type

    class Source(Enum):
        OTHER = 0
        AUDIO = 1
        MOVIE = 2
        SERIES = 3

        @staticmethod
        def from_name(name: str):
            """
            Returns a `Type.Source` by its name

            Args:
                name: Name of the value

            Returns:
                Type.Source: The corresponding type
            """
            for type in Type.Source:
                if type.name.lower() == name.lower():
                    return type


class Database(QtSql.QSqlDatabase):

    def __init__(self, database: QtSql.QSqlDatabase, auto_table=True):
        super().__init__(database)

        self._auto_table = auto_table

        # if `auto_table` is True it will be checked if every table exists and if not
        # it will be created
        if auto_table:
            tables = self.tables()
            if 'clip' not in tables:
                self.new_query().exec('CREATE TABLE clip (id int(10) DEFAULT NULL,'
                                      'source_id int(10) DEFAULT NULL,'
                                      'path text DEFAULT NULL,'
                                      'type smallint(1) DEFAULT NULL,'
                                      'name text DEFAULT NULL,'
                                      'description text DEFAULT NULL)')
                logger.debug('Created clip table')
            if 'item' not in tables:
                self.new_query().exec('CREATE TABLE item (id int(10) DEFAULT NULL,'
                                      'clip_id int(10) DEFAULT NULL,'
                                      'tag_id int(10) DEFAULT NULL)')
                logger.debug('Created item table')
            if 'source' not in tables:
                self.new_query().exec('CREATE TABLE source (id int(10) DEFAULT NULL,'
                                      'path text DEFAULT NULL,'
                                      'type smallint(1) DEFAULT NULL,'
                                      'name text DEFAULT NULL,'
                                      'season text DEFAULT NULL,'
                                      'episode text DEFAULt NULL,'
                                      'description text DEFAULT NULL)')
                logger.debug('Created source table')
            if 'tag' not in tables:
                self.new_query().exec('CREATE TABLE tag (id int(10) DEFAULT NULL,'
                                      'name text DEFAULT NULL,'
                                      'color char(7) DEFAULT "#c8c8c8")')  # <- will be needed in a future update
                logger.debug('Created tag table')

        # initializes the table classes and makes it public available
        self.Clip = self._Clip(self)
        self.Item = self._Item(self)
        self.Source = self._Source(self)
        self.Tag = self._Tag(self)

    class _Clip(_Table):
        """Class for the 'clip' table"""

        table_name = 'clip'

        def __init__(self, database):
            super().__init__(database, Database._Clip.table_name, Clip(database, None, 0, 0, "", Type.Clip.OTHER.value, "", ""))

        def add_clip(self, path: str, source_id: int = None, type: Union[Type.Clip, int] = Type.Clip.OTHER, name: str = None, description: str = None):
            """
            Adds a new clip to the database

            Args:
                path: Path of the clip
                source_id: Id of the clips source
                type: Type of the clip to add
                name: Name of the clip
                description: Description of the clip

            Returns:
                Clip: The new created clip

            Raises:
                TypeError: If `type_` is int and ``Type.Clip`` has no value like `type_`
            """

            if isinstance(type, int):
                type = Type.Clip(type)

            clip_id = self.highest_id() + 1

            if source_id is None:
                source_id = 0

            # adds the new clip to the table
            query = self.new_query()
            query.exec(f'INSERT INTO {self.table_name} (id, source_id, path, type, name, description)'
                       f'VALUES ({clip_id}, {source_id}, "{path}", {type.value}, "{name}", "{description}")')

            if self.validate_query(query):
                return Clip(self.database, None, clip_id, source_id, path, type.value, name, description)
            else:
                return Clip(self.database)

        def add_clips(self, clips: List[Dict[str, Union[str, int, Type.Clip]]]):
            """
            Adds a new clip to the database

            Args:
                clips: The clips to add. See `self.add_clip(...)` for more information about the elements

            Returns:
                List[Clip]: A list of the new created clips

            """
            keys = ['path', 'source_id', 'type', 'name', 'description']

            new_clips = []
            highest_id = self.highest_id()

            for i, clip in enumerate(clips):
                clip['id'] = highest_id + i + 1
                for key in keys:
                    # adds the non existing key to the dict
                    if key not in clip:
                        if key == 'path':
                            raise IOError(f'path must be specified in {clip}')
                        elif key in ('name', 'description'):
                            clip[key] = ''
                        elif key == 'source_id':
                            clip[key] = 0
                        elif key == 'type':
                            clip[key] = Type.Clip.OTHER
                    if key == 'type' and not isinstance(key, int):
                        clip[key] = Type.Clip(clip[key])
                new_clips.append(Clip(self.database, None, clip))

            query = self.new_query()
            # adds the new clips to the table
            query.exec(f'INSERT INTO {self.table_name} (id, source_id, path, type, name, description) VALUES '
                       f'({"), (".join([", ".join(self.make_sql_valid(clip["id"], [clip["source_id"], clip["path"], clip["type"].value, clip["name"], clip["description"]])) for clip in clips])})')
            self.validate_query(query, True)


    class _Item(_Table):
        """Class for the 'item' table"""

        table_name = 'item'

        def __init__(self, database):
            super().__init__(database, Database._Item.table_name, Item(database, None, 0, 0, 0))

        def add_item(self, clip_id, tag_id):
            """
            Adds a new item to the database

            Args:
                tag_id: Id of the tag
                clip_id: Id of the clip

            Returns:
                Item_ The new created item

            """
            item_id = self.highest_id() + 1

            # adds the new item to the table
            query = self.new_query()
            query.exec(f'INSERT INTO {self.table_name} (id, clip_id, tag_id)'
                       f'VALUES ({item_id}, {clip_id}, {tag_id})')

            if self.validate_query(query):
                return Item(self.database, None, item_id, clip_id, tag_id)
            else:
                return Item(self.database)

    class _Source(_Table):
        """Class for the 'source' table"""

        table_name = 'source'

        def __init__(self, database):
            super().__init__(database, Database._Source.table_name, Source(database, None, 0, "", Type.Source.OTHER.value, "", "", "", ""))

        def add_source(self, path_or_url: str, type: Union[Type.Source, int], name: str, description: str = None, season: str = None, episode: str = None):
            """
            Adds a new source to the database

            Args:
                path_or_url: Path or url to the source
                type: Type of the source to add
                name: Name of the source
                description: Description of the source
                season: Season of the source. Is only used if type_ is `Type.Source.SERIES`
                episode: Episode of the source. Is only used if type_ is `Type.Source.SERIES`

            Returns:
                Source: The new created source

            Raises:
                TypeError: If `type_` is int and ``Type.Source`` has no value like `type_`
            """
            if isinstance(type, int):
                type = Type.Source(type)
            if type is not Type.Source.SERIES:
                season = None
                episode = None

            source_id = self.highest_id() + 1

            # adds the new source to the table
            query = self.new_query()
            query.exec(f'INSERT INTO {self.table_name} (id, path, type, name, description, season, episode)'
                       f'VALUES ({source_id}, "{path_or_url}", {type.value}, "{name}", "{description}", "{season}", "{episode}")')

            if self.validate_query(query):
                return source_id
            else:
                return 0

    class _Tag(_Table):
        """Class for the 'tag' table"""

        table_name = 'tag'

        def __init__(self, database):
            super().__init__(database, Database._Tag.table_name, Tag(database, None, 0, ""))

        def add_tag(self, tag_name: str):
            """
            Adds a new tag to the database

            Args:
                tag_name: Name of the new tag

            Returns:
                Tag: The new created tag
            """
            tag_id = self.highest_id() + 1

            # adds the new tag to the database
            query = self.new_query()
            query.exec(f'INSERT INTO {self.table_name} (id, name)'
                       f'VALUES ({tag_id}, "{tag_name}")')

            if self.validate_query(query):
                return Tag(self.database, None, tag_id, tag_name)
            else:
                return Tag(self.database)

        def get_tags_by_clip_id(self, clip_id: int):
            """
            Receive all tags for a specific clip

            Args:
                clip_id: Id of the clip

            Returns:
                ClipRows: All tags from the given clip ids

            """
            query = self.new_query()
            query.exec(f'SELECT tag_id FROM {Item.table_name} WHERE clip_id={clip_id}')
            if self.validate_query(query):
                ids = []
                while query.next():
                    ids.append(str(query.value(0)))
                query = self.new_query()
                query.exec(f'SELECT * FROM {Tag.table_name} WHERE id IN ({", ".join(ids)})')
                if self.validate_query(query):
                    return TagRows(self.database, query=query)
            return TagRows(self.database)

    def new_query(self) -> QtSql.QSqlQuery:
        return QtSql.QSqlQuery(self)


class _Row:
    """Class where all information about a row in a table is stored"""

    def __init__(self, database: Database, query: QtSql.QSqlQuery = None, *args):
        self._database = database
        self.query = query

        # makes the id better available
        if args:
            self.id = args[0]
        else:
            self.query.next()
            self.id = query.value(0)


class Clip(_Row):
    """Class where all information about a row in the 'clip' table is stored"""

    table_name = 'clip'

    def __init__(self, database: Database, query: QtSql.QSqlQuery = None, *args):
        super().__init__(database, query, *args)

        if args:
            self.source_id: int = args[1]
            self.path: str = args[2]
            self.type = Type.Clip(args[3])
            self.name: str = args[4]
            self.description: str = args[5]
        else:
            self.source_id: int = self.query.value(1)
            self.path: str = self.query.value(2)
            self.type = Type.Clip(self.query.value(3))
            self.name: str = self.query.value(4)
            self.description: str = self.query.value(5)

    def __hash__(self):
        return hash((self.source_id, self.path, self.type, self.name, self.description))

    def get_source(self):
        """
        Get the clip's source

        Returns:
            Source: The clip's source
        """
        return self._database.Source.get(self.source_id)

    def get_tags(self):
        """
        Get all tags with which the clip is tagged

        Returns:
            TagRows: All tags from the clip
        """
        return self._database.Tag.get_tags_by_clip_id(self.id)

    def remove_tag(self, tag_id: int) -> None:
        """
        Removes a tag from the clip

        Args:
            tag_id: Id of the tag to remove
        """
        self._database.Item.delete(self._database.Item.get_by(clip_id=self.id, tag_id=tag_id).id)


class Item(_Row):
    """Class where all information about a row in the 'item' table is stored"""

    table_name = 'item'

    def __init__(self, database: Database, query: QtSql.QSqlQuery = None, *args):
        super().__init__(database, query, *args)

        if args:
            self.clip_id: int = args[1]
            self.tag_id: int = args[2]
        else:
            self.clip_id: int = self.query.value(1)
            self.tag_id: int = self.query.value(2)

    def __hash__(self):
        return hash((self.clip_id, self.tag_id))


class Source(_Row):
    """Class where all information about a row in the 'source' table is stored"""

    table_name = 'source'

    def __init__(self, database: Database, query: QtSql.QSqlQuery = None, *args):
        super().__init__(database, query, *args)

        if args:
            self.path: str = args[1]
            self.type = Type.Source(args[2])
            self.name: str = args[3]
            self.season: str = args[4]
            self.episode: str = args[5]
            self.description: str = args[6]
        else:
            self.path: str = self.query.value(1)
            self.type = Type.Source(self.query.value(2))
            self.name: str = self.query.value(3)
            self.season: str = self.query.value(4)
            self.episode: str = self.query.value(5)
            self.description: str = self.query.value(6)

    def __hash__(self):
        return hash((self.path, self.type, self.name, self.season, self.episode, self.description))

    def get_all_clips(self):
        """
        Get all clips using this source

        Returns:
            ClipRows: All clips using this source
        """
        return self._database.Clip.get_all_by(source_id=self.id)


class Tag(_Row):
    """Class where all information about a row in the 'tag' table is stored"""

    table_name = 'tag'

    def __init__(self, database: Database, query: QtSql.QSqlQuery = None, *args):
        super().__init__(database, query, *args)

        if args:
            self.name: str = args[1]
        else:
            self.name: str = self.query.value(1)

    def __hash__(self):
        return hash((self.name,))


class _Rows(list):
    """Class where all information about multiple rows in a table are stored in"""

    def __init__(self, database: Database, row_type: _Row.__class__, rows: List[_Row.__class__] = None, query: QtSql.QSqlQuery = None):
        super().__init__()
        self._database = database
        self.query = query

        if rows:
            self.extend(rows)
        else:
            if query:
                columns = query.record().count()
                # iterate through the query to get all values
                while query.next():
                    self.append(row_type(database, *[query.value(count) for count in range(-1, columns)]))
            else:
                raise ValueError('rows and query cannot be None at the same time')

    def __hash__(self):
        return hash((hash(item) for item in self))

    def new_query(self) -> QtSql.QSqlQuery:
        """
        Get a new sql query

        Returns:
            The sql query
        """
        return QtSql.QSqlQuery(self._database)


class ClipRows(_Rows):
    """Class where all information about multiple rows in the 'clip' table are stored in"""

    table_name = Clip.table_name

    def __init__(self, database: Database, rows: List[Clip] = None, query: QtSql.QSqlQuery = None):
        super().__init__(database, Clip, rows, query)

        self.table_name = Database._Clip.table_name


class ItemRows(_Rows):
    """Class where all information about multiple rows in the 'item' table are stored in"""

    table_name = Item.table_name

    def __init__(self, database: Database, rows: List[Item] = None, query: QtSql.QSqlQuery = None):
        super().__init__(database, Item, rows, query)

        self.table_name = Database._Item.table_name

    def get_clips(self) -> ClipRows:
        """
        Get all clips which have one of the here saved clip_ids as a id

        Returns:
            All corresponding clips
        """
        query = self.new_query()
        query.exec(f'SELECT * FROM {Database._Clip.table_name} WHERE id IN ({", ".join(item.tag_id for item in self)})')
        return ClipRows(self._database, query=query)


class SourceRows(_Rows):
    """Class where all information about multiple rows in the 'source' table are stored in"""

    table_name = Source.table_name

    def __init__(self, database: Database, rows: List[Source] = None, query: QtSql.QSqlQuery = None):
        super().__init__(database, Source, rows, query)

        self.table_name = Database._Source.table_name

    def get_tags(self):
        """
        Get all tags which have one of the here saved sources as source

        Returns:
            TagRows: All corresponding tags
        """
        query = self.new_query()
        query.exec(f'SELECT * FROM {Tag.table_name} WHERE source_id IN ({", ".join(str(item.id) for item in self)})')
        return TagRows(self._database, query=query)


class TagRows(_Rows):
    """Class where all information about multiple rows in the 'tag' table are stored in"""

    table_name = Tag.table_name

    def __init__(self, database: Database, rows: List[Tag] = None, query: QtSql.QSqlQuery = None):
        super().__init__(database, Tag, rows, query)

        self.table_name = Database._Tag.table_name

    def get_items(self) -> ItemRows:
        """
        Get all clips which have one of the here saved tags as a tag

        Returns:
            All corresponding clips
        """
        query = self.new_query()
        query.exec(f'SELECT clip_id FROM {Database._Item.table_name} WHERE tag_id IN ({", ".join(item.id for item in self)})')
        return ItemRows(self._database, query=query)
