from dreamutils.os import platform, Platform
from PyQt5.QtWidgets import QApplication
from typing import Any, Dict, Union, Mapping, overload, Iterable, Tuple
from pathlib import Path
import logging
import re
import sys
import urllib.error, urllib.request
import json


class AthnosPath:

    @staticmethod
    def athnos_path() -> Path:
        """Returns the path where all athnos stuff like clips or configurations are stored"""
        if platform() == Platform.WINDOWS:
            clip_path = Path.home().joinpath('AppData\\Roaming\\athnos')
        else:
            clip_path = Path.home().joinpath('.athnos/')

        return clip_path

    @classmethod
    def clips_path(cls) -> Path:
        """Returns the path where the clips should be stored"""
        return cls.athnos_path().joinpath('clips')

    @classmethod
    def config_path(cls) -> Path:
        """Path to the config file"""
        return cls.athnos_path().joinpath('config.json')

    @classmethod
    def logging_path(cls) -> Path:
        """Path to the log file"""
        return cls.athnos_path().joinpath('athnos.log')


class Config(dict):
    """Class to read / edit the configuration"""

    config_path = AthnosPath.config_path()

    def __init__(self, auto_write=True):
        self.auto_write = auto_write
        # creates the file if it doesn't exits
        if not self.config_path.is_file():
            super().__init__()
            self.config_path.parent.mkdir(parents=True, exist_ok=True)
            self.set_default()

        # loads the config file
        file = self.config_path.open('r')
        super().__init__(json.load(file))

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        if self.auto_write:
            self.write()

    def set_default(self) -> None:
        """Sets the default parameter of the file"""
        json.dump({'check_for_updates': True,
                   'update-notification': 'minor',
                   'updates_ignore': [],
                   'style': QApplication.style().objectName(),
                   'database': {'title': 'SQLite3', 'driver': 'QSQLITE', 'name': 'athnos.sqlite3', 'user': None, 'password': None, 'ip': None, 'port': None}},
                  open(self.config_path, 'w'), indent=4)

    def write(self) -> None:
        """Writes the config file"""
        json.dump(self, open(self.config_path, 'w'), indent=4)


def check_version(current_version: str) -> Union[Dict[str, Any], None]:
    """Checks if a new version was released"""
    from . import config, logger

    update_notification = config['update-notification']
    # checks if '-' is in `current-version`.
    # might be possible if version is x.y.z-beta or something else
    if '-' in current_version:
        internal_version, internal_pre_release = current_version.split('-')
        internal_version = internal_version.split('.')
    else:
        internal_version = current_version.split('.')

    if update_notification is not None:
        try:
            # fetch all release versions
            latest = urllib.request.urlopen('https://api.github.com/repos/ByteDream/athnos/tags')
            data = json.loads(latest.read().decode())

            notification = False

            try:
                notification_level = {'major': 0, 'minor': 1, 'patch': 2, 'all': 3}[update_notification]
                full_break = False

                for release in data:
                    # searches the version number and the eventual pre release
                    release_regex = re.search(r'^v(?P<version>(?:\d*\.){,2}\d*)-?(?P<pre_release>.*$)', release['name'])
                    version = release_regex.group('version').split('.')

                    if release['name'] in config['updates_ignore']:
                        continue

                    # if the version is not x.y.z
                    if len(version) == 2:
                        version.append('0')

                    if version > internal_version:
                        for i in range(0, len(version)):
                            if version[i] > internal_version[i]:
                                if notification_level >= i:
                                    notification = True
                            elif version[i] < internal_version[i]:
                                full_break = True
                    else:
                        break

                    if full_break or notification_level:
                        break
            except KeyError:
                pass

            if notification:
                return release

        except urllib.error.HTTPError as e:
            logger.warning(f'Could not get releases (Status: {e.getcode()}): {e.read().decode()}')
        except urllib.error.URLError as e:
            logger.warning(f'Could not get releases: {e.reason}. '
                           f'This error often occurs when a computer has no internet connection')


def get_logger(file_output=True, console_output=True, level=logging.DEBUG) -> logging.Logger:
    """
    Creates a new logger

    Args:
        file_output: If the logger should log to file or not
        console_output: If the logger should log to the console or not

    Returns:
        The new created logger
    """
    logger = logging.getLogger('athnos')
    if logger.hasHandlers():
        return logger

    # sets the logger output format
    formatter = logging.Formatter('[%(asctime)s] - %(levelname)s: %(message)s', '%H:%M:%S')
    # creates the logger handlers

    # configuring the logger
    logger.setLevel(level)

    if console_output:
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)
    if file_output:
        file_handler = logging.FileHandler(AthnosPath.logging_path())
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    return logger


def remove_prefix(string: str, prefix: str) -> str:
    """
    Removes the given prefix from the given string

    Args:
        string: string to cut the prefix off
        prefix: Prefix to cut off

    Returns:
        The string without the prefix
    """
    # uses the internal `removeprefix` method if the python version is higher or equal then 3.9
    if sys.version_info >= (3, 9):
        return string.removeprefix(prefix)
    elif string.startswith(prefix):
        return string[len(prefix):]
    # returns the string, if its not starting with the prefix
    else:
        return string
