from PyQt5 import QtGui, QtCore, QtWidgets
from typing import Any, Dict, List, Union
from . import resources


# https://github.com/baoboa/pyqt5/blob/master/examples/layouts/flowlayout.py
class QFlowLayout(QtWidgets.QLayout):
    def __init__(self, parent=None, margin=0, spacing=-1):
        super().__init__(parent)

        if parent is not None:
            self.setContentsMargins(margin, margin, margin, margin)

        self.setSpacing(spacing)

        self._items = []
        self.__pending_positions = {}

    def __del__(self):
        item = self.takeAt(0)
        while item:
            item = self.takeAt(0)

    def addItem(self, a0: QtWidgets.QLayoutItem) -> None:
        try:
            position = self.__pending_positions[a0.widget()]
            self._items.insert(position, a0)
            del self.__pending_positions[a0.widget()]
        except KeyError:
            self._items.append(a0)

    def addWidget(self, w: QtWidgets.QWidget, position: int = None) -> None:
        if position:
            self.__pending_positions[w] = position
        super().addWidget(w)

    def count(self):
        return len(self._items)

    def expandingDirections(self):
        return QtCore.Qt.Orientations(QtCore.Qt.Orientation(0))

    def itemAt(self, index: int) -> QtWidgets.QLayoutItem:
        if 0 <= index < len(self._items):
            return self._items[index]

        return None

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width):
        height = self._doLayout(QtCore.QRect(0, 0, width, 0), True)
        return height

    def minimumSize(self):
        size = QtCore.QSize()

        for item in self._items:
            size = size.expandedTo(item.minimumSize())

        margin, _, _, _ = self.getContentsMargins()

        size += QtCore.QSize(2 * margin, 2 * margin)
        return size

    def removeItem(self, a0: QtWidgets.QLayoutItem) -> None:
        a0.widget().deleteLater()

    def removeWidget(self, w: QtWidgets.QWidget) -> None:
        w.deleteLater()

    def setGeometry(self, rect):
        super().setGeometry(rect)
        self._doLayout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def takeAt(self, index: int) -> QtWidgets.QLayoutItem:
        if 0 <= index < len(self._items):
            return self._items.pop(index)

        return None

    def _doLayout(self, rect, testOnly):
        """This does the layout. Dont ask me how. Source: https://github.com/baoboa/pyqt5/blob/master/examples/layouts/flowlayout.py"""
        x = rect.x()
        y = rect.y()
        line_height = 0

        for item in self._items:
            wid = item.widget()
            space_x = self.spacing() + wid.style().layoutSpacing(
                QtWidgets.QSizePolicy.PushButton,
                QtWidgets.QSizePolicy.PushButton,
                QtCore.Qt.Horizontal)
            space_y = self.spacing() + wid.style().layoutSpacing(
                QtWidgets.QSizePolicy.PushButton,
                QtWidgets.QSizePolicy.PushButton,
                QtCore.Qt.Vertical)
            next_x = x + item.sizeHint().width() + space_x
            if next_x - space_x > rect.right() and line_height > 0:
                x = rect.x()
                y = y + line_height + space_y
                next_x = x + item.sizeHint().width() + space_x
                line_height = 0

            if not testOnly:
                item.setGeometry(QtCore.QRect(QtCore.QPoint(x, y), item.sizeHint()))

            x = next_x
            line_height = max(line_height, item.sizeHint().height())

        return y + line_height - rect.y()


class QTagEdit(QtWidgets.QScrollArea):
    """A tag based edit"""

    def __init__(self, parent: QtWidgets = None, tag_suggestions: List[str] = []):
        super().__init__(parent)

        # setup the ui stuff
        self.setWidgetResizable(True)

        self._main_widget = QtWidgets.QWidget()
        self._layout = QFlowLayout(self._main_widget)

        self._tag_input = QtWidgets.QLineEdit()
        # self._tag_input.setPlaceholderText('Type in a new tag and hit enter...')
        self._tag_input.setFixedWidth(10)
        self._tag_input.setStyleSheet('border: 0px')
        self._tag_input.setContentsMargins(0, 5, 0, 0)
        self._tag_input.keyReleaseEvent = self.__tag_input_key_release_event

        self._layout.addWidget(self._tag_input)

        self._tag_input.palette().color(QtGui.QPalette.Background)
        tag_input_color = self._tag_input.palette().color(QtGui.QPalette.Background)
        self.setStyleSheet(f'background-color: rgb({tag_input_color.red()}, {tag_input_color.green()}, {tag_input_color.blue()})')

        self.setWidget(self._main_widget)

        # setup all other things
        self.__font_calculator = QtGui.QFontMetrics(QtWidgets.QApplication.font())
        self.__tags = []
        self.__tag_suggestions = tag_suggestions
        self._tag_suggestions = False
        self._check_for_doubles = True

    def focusInEvent(self, a0: QtGui.QFocusEvent) -> None:
        """Sets the focus always to `self._tag_input`"""
        self._tag_input.setFocus()

    def addTag(self, text: str) -> bool:
        """
        Adds a new tag

        Args:
            text: The text of the new tag

        Returns:
            If the tag was added successfully
        """
        # if `self._check_for_doubles` is True, it checks if the tag already exists
        if self._check_for_doubles and text.lower() in (tag.lower() for tag in self.__tags):
            self.onDoubledTag(text)
            return False
        else:
            # a new tag
            tag = self.__QTagFrame(self, text)
            # tag.setStyleSheet('border: 0px; margin: 0px; padding: 0px')

            self.__tags.append(text)
            for tag_name in self.__tag_suggestions:
                # if the tag is in `self.__tag_suggestions` it will be removed from there
                if tag_name.lower() == text:
                    self.__tag_suggestions.remove(tag_name)
                    self.enableTagSuggestions(self._tag_suggestions)
                    break
            # insert the tag before the line edit
            self._layout.addWidget(tag, -1)
            return True

    def clear(self, input=True) -> None:
        """
        Clears all tags

        Args:
            input: If True, the current text in the line edit will be cleared as well
        """
        for i in range(self._layout.count()):
            widget = self._layout.itemAt(i).widget()
            if type(widget) == QtWidgets.QLineEdit and input:
                widget.clear()
            else:
                self.removeTag(widget)

    def enableCheckForDoubles(self, check_for_doubles) -> None:
        """
        Enables if a new tag, when its going to be added, should be checked if it already exists

        Args:
            check_for_doubles: True if double checking should be active, False if not
        """
        self._check_for_doubles = check_for_doubles

    def enableTagSuggestions(self, tag_suggestions: bool) -> None:
        """
        Enables whenever a new tag is typed in that suggestions from `self.__tag_suggestions(...)` should be showing or not.
        They can be added on the class initialization or via `setTagSuggestions`

        Args:
            tag_suggestions: If tag suggestions should be active or not
        """
        if tag_suggestions:
            # sets the completer for the line edit
            completer = QtWidgets.QCompleter(self.__tag_suggestions, self)
            completer.setCaseSensitivity(QtCore.Qt.CaseSensitive)
            self._tag_input.setCompleter(completer)
        else:
            self._tag_input.setCompleter(None)

    def onDoubledTag(self, text: str) -> None:
        """
        This method gets called if `self._check_for_doubles` is True (can be set via `enableCheckForDoubles(...)`)
        and a new tag which already exists are going to be added.
        This method is actually there to display an error message

        Args:
            text: The text of the new tag
        """
        button = QtWidgets.QMessageBox()

        # if `self.tag_input.keyReleaseEvent`is not overridden, it would trigger itself when the enter key is pressed
        # to close the popup and open it again. idk why
        def reset(a0: QtGui.QKeyEvent):
            if a0.key() in (QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter):
                button.destroy()
                self._tag_input.keyReleaseEvent = self.__tag_input_key_release_event

        self._tag_input.keyReleaseEvent = reset

        # shows the warning
        button.warning(self, 'Tag already exists', f'The tag {text} already exists')

    def removeTag(self, tag) -> None:
        """
        Removes a tag

        Args:
            tag (QTagEdit.__QTagFrame): The tag to remove
        """
        self.__tags.remove(tag.text())
        self._layout.removeWidget(tag)

    def setTags(self, tags: List[str]) -> None:
        """
        Replaces all current tags with tag from the `tags` argument

        Args:
            tags: The new tags to be set
        """
        self.clear()
        for tag in tags:
            self.addTag(tag)

    def setTagSuggestions(self, suggenstions: List[str]) -> None:
        """
        Sets the tag suggestions. They will be used if `self._tag_suggestions` is True (can be set via `enableTagSuggestions(...)`)
        and will be shown if a new tag is typed in

        Args:
            suggenstions: The new tag suggestions
        """
        self.__tag_suggestions = suggenstions
        self.enableTagSuggestions(self._tag_suggestions)

    def tags(self) -> Union[List[str], List]:
        """
        Returns all tag names

        Returns:
            All tag names
        """
        return self.__tags

    def __tag_input_key_release_event(self, a0: QtGui.QKeyEvent) -> None:
        """
        The `keyReleaseEvent(...)` of the line edit. Whenever return / enter is pressed, the current text in the line edit
        will be added as new tag. It also expands the width of the line edit if the text in it is over an specific limit
        """
        if a0.key() in (QtCore.Qt.Key_Enter, QtCore.Qt.Key_Return):
            # adds the tag
            if self.addTag(self._tag_input.text()):
                self._tag_input.clear()
                self._tag_input.setFixedWidth(10)
            return
        elif a0.key() in (QtCore.Qt.Key_Backspace, QtCore.Qt.Key_Delete):
            # calculates the current line edit text width
            width = self.__font_calculator.width(self._tag_input.text())
            if width + 10 < self._tag_input.width():
                self._tag_input.setFixedWidth(width + 10)
        else:
            # calculates the current line edit text width
            width = self.__font_calculator.width(self._tag_input.text())
            # this resizes the tag input if the text in it will be longer than it's width.
            # not the best way, but it does what it does
            if width + 20 > self._tag_input.width():
                self._tag_input.setFixedWidth(width + 20)

    class __QTagFrame(QtWidgets.QFrame):
        """The tag class for the QTagEdit tags"""

        def __init__(self, parent, text: str):
            super().__init__()

            # setup the ui stuff
            self.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed))
            self.setLayout(QtWidgets.QHBoxLayout())
            self.setContentsMargins(3, 0, 3, 0)

            # the tag name label
            self._name = QtWidgets.QLabel()
            self._name.setText(text)
            self._name.setStyleSheet('background: transparent')

            # the tag delete button
            self._delete_button = QtWidgets.QPushButton()
            self._delete_button.setStyleSheet('QPushButton{background: transparent; image: url(:/tags/delete); border: 0px}'
                                              'QPushButton:hover {image: url(:/tags/delete_hover)}')
            self._delete_button.clicked.connect(self.onDeleteButtonClick)

            self.layout().addWidget(self._name)
            self.layout().addWidget(self._delete_button)

            # setup all other things
            self.__parent = parent
            self._text = text

        def onDeleteButtonClick(self) -> None:
            """This will get triggered if the delete button on a tag is pressed"""
            self.__parent.removeTag(self)
            self.__parent._tag_input.setFocus()

        def paintEvent(self, a0: QtGui.QPaintEvent) -> None:
            """Styles the tag"""
            painter = QtGui.QPainter(self)
            painter.setRenderHint(QtGui.QPainter.Antialiasing)
            painter.setPen(QtGui.QPen(QtGui.QColor(200, 200, 200, 255), 1, QtCore.Qt.SolidLine))
            painter.setBrush(QtGui.QBrush(QtGui.QColor(200, 200, 200, 255), QtCore.Qt.SolidPattern))

            # draws the tag 'filling'
            painter.drawRoundedRect(0, 0, self.width() - 5, self.height(), self.height() / 2, self.height() / 2)

        def text(self) -> str:
            """
            Returns the current tag text

            Returns:
                The tag text

            """
            return self._text


class QTableSortFilterProxyModel(QtCore.QSortFilterProxyModel):
    """The normal `QtCore.QSortFilterProxyModel` class but with table / multiple column support"""

    def __init__(self, overwrite_filters=False):
        super().__init__()

        self.overwrite_filters = overwrite_filters

        # all filters
        self.filters: Dict[int, QtCore.QRegExp] = {}

    def clearFilter(self) -> None:
        """Removes all filter"""
        self.filters = {}

    # https://stackoverflow.com/questions/60425164/adding-a-checkbox-column-to-a-custom-qsortfilterproxymodel
    #def data(self, index, role: int = 0):
    #    return QtCore.Qt.Checked

    def setRegExpFilter(self, column: int, regExp: QtCore.QRegExp) -> None:
        """
        Adds a new regex filter to a specific column

        Args:
            column: The column to add the filter to
            regExp: The regex filter

        Raises:
            KeyError: If the column has already a filter
        """
        if column in self.filters and not self.overwrite_filters:
            raise KeyError('The column already has a filter')
        self.filters[column] = regExp

    def removeRegExpFilter(self, column: int) -> None:
        """
        Removes the regex filter for a specific column

        Args:
            column: The column to remove the regex filter from

        Raises:
            KeyError: If the column has no filter
        """
        if column not in self.filters:
            raise KeyError('The column has no filter')
        del self.filters[column]

    def filterAcceptsRow(self, source_row: int, source_parent: QtCore.QModelIndex) -> bool:
        # if one or more filters are set
        if self.filters:
            model = self.sourceModel()
            # loops through the given source model and checks if a regex filter is installed for one column
            for i in range(model.columnCount()):
                # if the current column has a filter and it matches with the index data / text
                if i in self.filters and self.filters[i].indexIn(str(model.data(model.index(source_row, i)))) != -1:
                    return True
            return False
        else:
            # if there are no filters, the normal `filterAcceptsRow(...)` gets called
            return super().filterAcceptsRow(source_row, source_parent)
