from pathlib import Path
from PyQt5 import QtCore, QtGui, QtWidgets
from . import _DatabaseWindow
import filetype
from ..database import Type
from .. import logger
from .QCustomObject import QTagEdit


class _Add(_DatabaseWindow):
    """Base class for every add window"""

    def __init__(self, window):
        super().__init__(window)

        old_key_press_event = self.window.keyPressEvent

        def new_key_press_event(a0: QtGui.QKeyEvent):
            if a0.key() in (QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter) and not self.file_or_url_edit.hasFocus():
                return
            else:
                old_key_press_event(a0)

        # the normal key press event would open the file choose dialog / press the file_chooser button
        # everytime enter his when a ``QLineEdit`` is focused
        self.window.keyPressEvent = new_key_press_event

    def after_setup_ui(self) -> None:
        # file chooser
        self.file_or_url_edit.textChanged.connect(self.on_text_changed)
        self.file_chooser.clicked.connect(self.on_choose_file)

        # save / new buttons
        self.save_button.clicked.connect(self.on_save)
        self.new_button.clicked.connect(self.on_new)

    def on_text_changed(self, value: str) -> None:
        """This will get called if the file or url line edit text is changing"""
        pass

    def on_choose_file(self) -> None:
        """Opens a file choose dialog"""
        # all unique media file types
        all_preferred_types = [extension.extension for clip_type in [filetype.audio_matchers, filetype.image_matchers, filetype.video_matchers] for extension in clip_type]
        file = QtWidgets.QFileDialog.getOpenFileName(self.window, directory=str(Path().home()),
                                                     filter=f'Preferred media file(*.{" *.".join(all_preferred_types)});;All files(*)')[0]
        if file:
            # checks the file type and sets the index if the type box after the type
            match = filetype.guess(file)
            if match in filetype.audio_matchers:
                self.type_box.setCurrentIndex(0)
            elif match in filetype.image_matchers:
                self.type_box.setCurrentIndex(1)
            elif match in filetype.video_matchers:
                self.type_box.setCurrentIndex(2)
            else:
                self.type_box.setCurrentIndex(3)
            # changes the file or url text edit text
            self.file_or_url_edit.setText(file)

    def on_save(self) -> None:
        """This will get called if the 'Save' button is clicked"""
        pass

    def on_new(self) -> None:
        """This will get called if the 'New' button is clicked"""
        pass


class AddClip(_Add):

    def __init__(self, window):
        super().__init__(window)

        # all tags which are currently in the database
        self._all_tags = [tag.name for tag in self.database.Tag.get_all()]
        self._source_id = 0

    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(400, 469)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.add_clip = QtWidgets.QLabel(Form)
        self.add_clip.setAlignment(QtCore.Qt.AlignCenter)
        self.add_clip.setObjectName("add_clip")
        self.verticalLayout_2.addWidget(self.add_clip)
        self.name_description_frame = QtWidgets.QFrame(Form)
        self.name_description_frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.name_description_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.name_description_frame.setObjectName("name_description_frame")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.name_description_frame)
        self.verticalLayout.setObjectName("verticalLayout")
        self.file_or_url_hlayout = QtWidgets.QHBoxLayout()
        self.file_or_url_hlayout.setObjectName("file_or_url_hlayout")
        self.file_or_url_edit = QtWidgets.QLineEdit(self.name_description_frame)
        self.file_or_url_edit.setText("")
        self.file_or_url_edit.setClearButtonEnabled(False)
        self.file_or_url_edit.setObjectName("file_or_url_edit")
        self.file_or_url_hlayout.addWidget(self.file_or_url_edit)
        self.file_chooser = QtWidgets.QPushButton(self.name_description_frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.file_chooser.sizePolicy().hasHeightForWidth())
        self.file_chooser.setSizePolicy(sizePolicy)
        self.file_chooser.setSizeIncrement(QtCore.QSize(0, 0))
        self.file_chooser.setCheckable(False)
        self.file_chooser.setObjectName("file_chooser")
        self.file_or_url_hlayout.addWidget(self.file_chooser)
        self.verticalLayout.addLayout(self.file_or_url_hlayout)
        self.line = QtWidgets.QFrame(self.name_description_frame)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.name_hlayout = QtWidgets.QHBoxLayout()
        self.name_hlayout.setObjectName("name_hlayout")
        self.name_label = QtWidgets.QLabel(self.name_description_frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.name_label.sizePolicy().hasHeightForWidth())
        self.name_label.setSizePolicy(sizePolicy)
        self.name_label.setMinimumSize(QtCore.QSize(70, 0))
        self.name_label.setBaseSize(QtCore.QSize(0, 0))
        self.name_label.setObjectName("name_label")
        self.name_hlayout.addWidget(self.name_label)
        self.name_edit = QtWidgets.QLineEdit(self.name_description_frame)
        self.name_edit.setStyleSheet("")
        self.name_edit.setReadOnly(True)
        self.name_edit.setClearButtonEnabled(False)
        self.name_edit.setObjectName("name_edit")
        self.name_hlayout.addWidget(self.name_edit)
        self.verticalLayout.addLayout(self.name_hlayout)
        self.description_hlayout = QtWidgets.QHBoxLayout()
        self.description_hlayout.setObjectName("description_hlayout")
        self.description_label = QtWidgets.QLabel(self.name_description_frame)
        self.description_label.setMinimumSize(QtCore.QSize(70, 0))
        self.description_label.setBaseSize(QtCore.QSize(0, 0))
        self.description_label.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.description_label.setObjectName("description_label")
        self.description_hlayout.addWidget(self.description_label)
        self.description_edit = QtWidgets.QTextEdit(self.name_description_frame)
        self.description_edit.setEnabled(True)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.description_edit.sizePolicy().hasHeightForWidth())
        self.description_edit.setSizePolicy(sizePolicy)
        self.description_edit.setReadOnly(True)
        self.description_edit.setObjectName("description_edit")
        self.description_hlayout.addWidget(self.description_edit)
        self.verticalLayout.addLayout(self.description_hlayout)
        self.type_hlayout = QtWidgets.QHBoxLayout()
        self.type_hlayout.setObjectName("type_hlayout")
        self.type_label = QtWidgets.QLabel(self.name_description_frame)
        self.type_label.setMinimumSize(QtCore.QSize(70, 0))
        self.type_label.setBaseSize(QtCore.QSize(0, 0))
        self.type_label.setObjectName("type_label")
        self.type_hlayout.addWidget(self.type_label)
        self.type_box = QtWidgets.QComboBox(self.name_description_frame)
        self.type_box.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.type_box.sizePolicy().hasHeightForWidth())
        self.type_box.setSizePolicy(sizePolicy)
        self.type_box.setFocusPolicy(QtCore.Qt.WheelFocus)
        self.type_box.setObjectName("type_box")
        self.type_box.addItem("")
        self.type_box.addItem("")
        self.type_box.addItem("")
        self.type_box.addItem("")
        self.type_hlayout.addWidget(self.type_box)
        self.verticalLayout.addLayout(self.type_hlayout)
        self.line_3 = QtWidgets.QFrame(self.name_description_frame)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.verticalLayout.addWidget(self.line_3)
        self.tags_hlayout = QtWidgets.QHBoxLayout()
        self.tags_hlayout.setObjectName("tags_hlayout")
        self.tags_label = QtWidgets.QLabel(self.name_description_frame)
        self.tags_label.setMinimumSize(QtCore.QSize(70, 0))
        self.tags_label.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.tags_label.setObjectName("tags_label")
        self.tags_hlayout.addWidget(self.tags_label)
        self.tags_edit = QtWidgets.QScrollArea(self.name_description_frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tags_edit.sizePolicy().hasHeightForWidth())
        self.tags_edit.setSizePolicy(sizePolicy)
        self.tags_edit.setWidgetResizable(True)
        self.tags_edit.setObjectName("tags_edit")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 290, 83))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.tags_edit.setWidget(self.scrollAreaWidgetContents)
        self.tags_hlayout.addWidget(self.tags_edit)
        self.verticalLayout.addLayout(self.tags_hlayout)
        self.line_2 = QtWidgets.QFrame(self.name_description_frame)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.verticalLayout.addWidget(self.line_2)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.choose_source_button = QtWidgets.QPushButton(self.name_description_frame)
        self.choose_source_button.setEnabled(False)
        self.choose_source_button.setMaximumSize(QtCore.QSize(100, 16777215))
        self.choose_source_button.setObjectName("choose_source_button")
        self.horizontalLayout.addWidget(self.choose_source_button)
        self.choose_source_label = QtWidgets.QLabel(self.name_description_frame)
        self.choose_source_label.setEnabled(False)
        self.choose_source_label.setObjectName("choose_source_label")
        self.horizontalLayout.addWidget(self.choose_source_label)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.verticalLayout_2.addWidget(self.name_description_frame)
        self.save_new_hlayout = QtWidgets.QHBoxLayout()
        self.save_new_hlayout.setObjectName("save_new_hlayout")
        self.save_button = QtWidgets.QPushButton(Form)
        self.save_button.setEnabled(False)
        self.save_button.setFlat(False)
        self.save_button.setObjectName("save_button")
        self.save_new_hlayout.addWidget(self.save_button)
        self.new_button = QtWidgets.QPushButton(Form)
        self.new_button.setCheckable(False)
        self.new_button.setChecked(False)
        self.new_button.setAutoRepeat(False)
        self.new_button.setAutoExclusive(False)
        self.new_button.setObjectName("new_button")
        self.save_new_hlayout.addWidget(self.new_button)
        self.verticalLayout_2.addLayout(self.save_new_hlayout)

        self.retranslateUi(Form)
        self.type_box.setCurrentIndex(3)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.add_clip.setText(_translate("Form", "Add clip"))
        self.file_or_url_edit.setPlaceholderText(_translate("Form", "File or url..."))
        self.file_chooser.setText(_translate("Form", "Choose file"))
        self.name_label.setText(_translate("Form", "Name"))
        self.description_label.setText(_translate("Form", "Description"))
        self.type_label.setText(_translate("Form", "Type"))
        self.type_box.setItemText(0, _translate("Form", "Audio"))
        self.type_box.setItemText(1, _translate("Form", "Image"))
        self.type_box.setItemText(2, _translate("Form", "Video"))
        self.type_box.setItemText(3, _translate("Form", "Other"))
        self.tags_label.setText(_translate("Form", "Tags"))
        self.choose_source_button.setText(_translate("Form", "Select source"))
        self.choose_source_label.setText(_translate("Form", "No source selected"))
        self.save_button.setText(_translate("Form", "Save"))
        self.save_button.setShortcut(_translate("Form", "Ctrl+S"))
        self.new_button.setText(_translate("Form", "New"))

    def after_setup_ui(self) -> None:
        super().after_setup_ui()

        # replaces the `self.tags_edit` scroll area with an `QCustomObject.QTagEdit`
        self.tags_edit.deleteLater()
        self.tags_edit.hide()
        self.tags_edit = QTagEdit()
        # configures the tage edit
        self.tags_edit.setEnabled(False)
        self.tags_edit.enableTagSuggestions(True)
        self.tags_edit.setTagSuggestions([tag.name for tag in self.database.Tag.get_all()])
        self.tags_hlayout.addWidget(self.tags_edit)

        self.choose_source_button.clicked.connect(self.on_select_source)

    def on_text_changed(self, text: str) -> None:
        if text.strip() == '':
            is_value = False
        else:
            is_value = True

        self.name_edit.setReadOnly(not is_value)
        self.description_edit.setReadOnly(not is_value)
        self.type_box.setEnabled(is_value)
        self.tags_edit.setEnabled(is_value)
        self.choose_source_button.setEnabled(is_value)
        self.choose_source_label.setEnabled(is_value)
        self.save_button.setEnabled(is_value)

    def on_select_source(self) -> None:
        """Opens a new `source.ShowSource` window to select the clips source"""

        # this is imported here, because it would cause a circular import if it's imported above
        from .source import ShowSource

        source = self.database.Source.get(ShowSource(QtWidgets.QDialog(self.window, QtCore.Qt.WindowSystemMenuHint)).select())
        self._source_id = source.id
        self.choose_source_label.setText(source.name)

    def on_save(self) -> None:
        if not self._source_id:
            reply = QtWidgets.QMessageBox.question(self.window,
                                                   'No source is given', 'No source for the clip is given. Are you sure to proceed?',
                                                   buttons=QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
            if reply == QtWidgets.QMessageBox.No:
                return

        type = Type.Clip.from_name(self.type_box.currentText().lower())
        if not type:
            return

        id = self.database.Clip.add_clip(self.file_or_url_edit.text(), self._source_id, type.value, self.name_edit.text(), self.description_edit.toPlainText()).id
        for tag in self.tags_edit.tags():
            if tag in self._all_tags:
                self.database.Item.add_item(id, self.database.Tag.get_by(name=tag).id)
            else:
                self.database.Item.add_item(id, self.database.Tag.add_tag(tag).id)
                logger.debug(f'Added new tag - Name: {tag}')

        logger.debug(f'Added new clip - file / url: "{self.file_or_url_edit.text()}", source id: "{self._source_id}", type: "{type.value}", '
                     f'name: "{self.name_edit.text()}", description: "{self.description_edit.toPlainText()}", tags: {", ".join(self.tags_edit.tags())}')

        self.on_new()

    def on_new(self) -> None:
        self.file_or_url_edit.setText('')
        self.name_edit.setText('')
        self.on_text_changed('')
        self.type_box.setCurrentIndex(3)
        self.tags_edit.clear()


class AddSource(_Add):

    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(452, 400)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setObjectName("verticalLayout")
        self.add_source_label = QtWidgets.QLabel(Form)
        self.add_source_label.setAlignment(QtCore.Qt.AlignCenter)
        self.add_source_label.setObjectName("add_source_label")
        self.verticalLayout.addWidget(self.add_source_label)
        self.frame = QtWidgets.QFrame(Form)
        self.frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.frame)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.file_or_url_hlayout = QtWidgets.QHBoxLayout()
        self.file_or_url_hlayout.setObjectName("file_or_url_hlayout")
        self.file_or_url_edit = QtWidgets.QLineEdit(self.frame)
        self.file_or_url_edit.setText("")
        self.file_or_url_edit.setObjectName("file_or_url_edit")
        self.file_or_url_hlayout.addWidget(self.file_or_url_edit)
        self.file_chooser = QtWidgets.QPushButton(self.frame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.file_chooser.sizePolicy().hasHeightForWidth())
        self.file_chooser.setSizePolicy(sizePolicy)
        self.file_chooser.setSizeIncrement(QtCore.QSize(0, 0))
        self.file_chooser.setObjectName("file_chooser")
        self.file_or_url_hlayout.addWidget(self.file_chooser)
        self.verticalLayout_2.addLayout(self.file_or_url_hlayout)
        self.line_3 = QtWidgets.QFrame(self.frame)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.verticalLayout_2.addWidget(self.line_3)
        self.type_hlayout = QtWidgets.QHBoxLayout()
        self.type_hlayout.setObjectName("type_hlayout")
        self.type_label = QtWidgets.QLabel(self.frame)
        self.type_label.setMinimumSize(QtCore.QSize(70, 0))
        self.type_label.setObjectName("type_label")
        self.type_hlayout.addWidget(self.type_label)
        self.type_box = QtWidgets.QComboBox(self.frame)
        self.type_box.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.type_box.sizePolicy().hasHeightForWidth())
        self.type_box.setSizePolicy(sizePolicy)
        self.type_box.setObjectName("type_box")
        self.type_box.addItem("")
        self.type_box.addItem("")
        self.type_box.addItem("")
        self.type_box.addItem("")
        self.type_hlayout.addWidget(self.type_box)
        self.verticalLayout_2.addLayout(self.type_hlayout)
        self.from_hlayout = QtWidgets.QHBoxLayout()
        self.from_hlayout.setObjectName("from_hlayout")
        self.name_label = QtWidgets.QLabel(self.frame)
        self.name_label.setMinimumSize(QtCore.QSize(70, 0))
        self.name_label.setObjectName("name_label")
        self.from_hlayout.addWidget(self.name_label)
        self.name_edit = QtWidgets.QLineEdit(self.frame)
        self.name_edit.setReadOnly(True)
        self.name_edit.setObjectName("name_edit")
        self.from_hlayout.addWidget(self.name_edit)
        self.season_label = QtWidgets.QLabel(self.frame)
        self.season_label.setObjectName("season_label")
        self.from_hlayout.addWidget(self.season_label)
        self.season_edit = QtWidgets.QLineEdit(self.frame)
        self.season_edit.setMaximumSize(QtCore.QSize(50, 16777215))
        self.season_edit.setReadOnly(True)
        self.season_edit.setObjectName("season_edit")
        self.from_hlayout.addWidget(self.season_edit)
        self.episode_label = QtWidgets.QLabel(self.frame)
        self.episode_label.setObjectName("episode_label")
        self.from_hlayout.addWidget(self.episode_label)
        self.episode_edit = QtWidgets.QLineEdit(self.frame)
        self.episode_edit.setMaximumSize(QtCore.QSize(50, 16777215))
        self.episode_edit.setReadOnly(True)
        self.episode_edit.setObjectName("episode_edit")
        self.from_hlayout.addWidget(self.episode_edit)
        self.verticalLayout_2.addLayout(self.from_hlayout)
        self.description_hlayout = QtWidgets.QHBoxLayout()
        self.description_hlayout.setObjectName("description_hlayout")
        self.description_label = QtWidgets.QLabel(self.frame)
        self.description_label.setMinimumSize(QtCore.QSize(70, 0))
        self.description_label.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.description_label.setObjectName("description_label")
        self.description_hlayout.addWidget(self.description_label)
        self.description_edit = QtWidgets.QTextEdit(self.frame)
        self.description_edit.setReadOnly(True)
        self.description_edit.setObjectName("description_edit")
        self.description_hlayout.addWidget(self.description_edit)
        self.verticalLayout_2.addLayout(self.description_hlayout)
        self.verticalLayout.addWidget(self.frame)
        self.save_new_hlayout = QtWidgets.QHBoxLayout()
        self.save_new_hlayout.setObjectName("save_new_hlayout")
        self.save_button = QtWidgets.QPushButton(Form)
        self.save_button.setEnabled(False)
        self.save_button.setObjectName("save_button")
        self.save_new_hlayout.addWidget(self.save_button)
        self.new_button = QtWidgets.QPushButton(Form)
        self.new_button.setObjectName("new_button")
        self.save_new_hlayout.addWidget(self.new_button)
        self.verticalLayout.addLayout(self.save_new_hlayout)

        self.retranslateUi(Form)
        self.type_box.setCurrentIndex(2)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.add_source_label.setText(_translate("Form", "Add source"))
        self.file_or_url_edit.setPlaceholderText(_translate("Form", "File or url..."))
        self.file_chooser.setText(_translate("Form", "Choose file"))
        self.file_chooser.setShortcut(_translate("Form", "Ctrl+O"))
        self.type_label.setText(_translate("Form", "Type"))
        self.type_box.setItemText(0, _translate("Form", "Audio"))
        self.type_box.setItemText(1, _translate("Form", "Movie"))
        self.type_box.setItemText(2, _translate("Form", "Series"))
        self.type_box.setItemText(3, _translate("Form", "Other"))
        self.name_label.setText(_translate("Form", "From"))
        self.season_label.setText(_translate("Form", "Season"))
        self.episode_label.setText(_translate("Form", "Episode"))
        self.description_label.setText(_translate("Form", "Description"))
        self.save_button.setText(_translate("Form", "Save"))
        self.save_button.setShortcut(_translate("Form", "Ctrl+S"))
        self.new_button.setText(_translate("Form", "New"))
        self.new_button.setShortcut(_translate("Form", "Ctrl+N"))

    def after_setup_ui(self) -> None:
        super().after_setup_ui()

        self.type_box.currentIndexChanged.connect(self.on_type_changed)

    def on_text_changed(self, text: str):
        if text.strip() == '':
            is_value = False
        else:
            is_value = True

        self.name_edit.setReadOnly(not is_value)
        self.season_edit.setReadOnly(not is_value)
        self.episode_edit.setReadOnly(not is_value)
        self.description_edit.setReadOnly(not is_value)
        self.type_box.setEnabled(is_value)
        self.save_button.setEnabled(is_value)

    def on_type_changed(self, index: int) -> None:
        """Modifies the name / from, season and episode line edits, based on the `index`"""
        if index in [0, 1, 3]:  # audio / movie
            # hides the season and episode label / line edit and sets the text of the name / from label to 'Name'
            self.name_label.setText('Name')
            self.season_label.hide()
            self.season_edit.hide()
            self.episode_label.hide()
            self.episode_edit.hide()
        elif index == 2:  # series
            # shows the season and episode label / line edit and sets the text of the name / from label to 'From'
            self.name_label.setText('From')
            self.season_label.show()
            self.season_edit.show()
            self.episode_label.show()
            self.episode_edit.show()

    def on_save(self) -> None:
        type = Type.Source.from_name(self.type_box.currentText().lower())
        if not type:
            return

        file_or_url = self.file_or_url_edit.text()
        name = self.name_edit.text()
        description = self.description_edit.toPlainText()
        season = self.season_edit.text()
        episode = self.episode_edit.text()

        # it checks if some of the data is already in the database
        if type == Type.Source.SERIES and self.database.Source.has(type=type.value, name=name, season=season, episode=episode):
            has_same = True
        elif type != Type.Source.SERIES and self.database.Source.has(name=name):
            has_same = True
        else:
            has_same = False

        if has_same:
            # if some of the data were in the database it checks if the exact same data is in the database
            if self.database.Source.has(path=file_or_url, type=type.value, name=name, description=description, season=season, episode=episode):
                QtWidgets.QMessageBox.warning(self.window, 'The source already exist', 'A source with the exact same data (name, description, type, etc.) already exist')
                return
            elif type == Type.Source.SERIES:
                proceed = QtWidgets.QMessageBox.question(self.window, 'The source may already exist', 'A source with some same data (name, season and episode at least) already exist. '
                                                                                                      'Do you want to proceed?', QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
                if not proceed:
                    return

        self.database.Source.add_source(self.file_or_url_edit.text(), type.value, self.name_edit.text(), self.description_edit.toPlainText(), self.season_edit.text(), self.episode_edit.text())
        logger.debug(f'Added new source - file / url: "{self.file_or_url_edit.text()}", name: "{self.name_edit.text()}", description: "{self.description_edit.toPlainText()}", '
                     f'season: "{self.season_edit.text()}", episode: "{self.episode_edit.text()}"')
        self.on_new()

    def on_new(self) -> None:
        self.file_or_url_edit.setText('')
        self.type_box.setCurrentIndex(2)
        self.name_edit.setText('')
        self.season_edit.setText('')
        self.episode_edit.setText('')
        self.description_edit.setText('')
