from PyQt5 import QtWidgets, QtSql, QtCore, QtGui
from .window import Window
from ..database import Database
from .. import logger


class _ImportExport:

    def __init__(self, main_window: Window, database_file: str):
        """
        Args:
            main_window Window: The instance of the main window
        """
        self.main_window = main_window

        self.old_database = self.main_window.database
        self.old_model = self.main_window.model

        db = QtSql.QSqlDatabase('QSQLITE')
        db.setDatabaseName(database_file)

        # tries to open a database connection
        if not db.open():
            raise IOError(db.lastError().text())

        self.main_window.window.setWindowTitle(f'Athnos [{self.__class__.__name__}]')
        self.main_window.window.setStyleSheet('#main_splitter {border: 1px solid #08F7FE;}')

        self.main_window.database = Database(db, False)
        self.main_window.load_model()

        self.extra_button_hlayout = QtWidgets.QHBoxLayout()
        self.cancel_button = QtWidgets.QPushButton('Cancel')
        self.cancel_button.clicked.connect(lambda: self.on_cancel())

        self.extra_button_hlayout.addWidget(self.cancel_button)

        self.main_window.verticalLayout.addLayout(self.extra_button_hlayout)

    def on_cancel(self) -> None:
        """Gets called if the cancel button is pressed"""
        result = QtWidgets.QMessageBox.warning(self.main_window.window, 'Cancel', 'Do you really want to cancel? All changes <strong>won\'t<strong> be saved!',
                                               QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel, QtWidgets.QMessageBox.Cancel)
        if result == QtWidgets.QMessageBox.Ok:
            self.reset()
        else:
            return

    def reset(self) -> None:
        """Resets the modified main window"""
        # deletes all widgets in the added layout
        for i in range(self.extra_button_hlayout.count()):
            self.extra_button_hlayout.itemAt(i).widget().deleteLater()

        # removes the added layout from the main window
        for i in range(self.main_window.verticalLayout.count()):
            item = self.main_window.verticalLayout.itemAt(i)
            if item.layout() == self.extra_button_hlayout:
                self.main_window.verticalLayout.removeItem(item)
                break

        self.main_window.window.setStyleSheet('')
        self.main_window.window.setWindowTitle('Athnos')
        self.main_window.database = self.old_database
        self.main_window.load_model()


class Import(_ImportExport):

    def __init__(self, main_window: Window):
        file = QtWidgets.QFileDialog.getOpenFileName(filter='sqlite3 database (*.sqlite3 *.sqlite);;zip file (*.zip);;All files (*.*)')[0]
        if not file:
            self.main_window.close()
            return
        super().__init__(main_window, file)

        self.import_button = QtWidgets.QPushButton('Import')
        self.import_button.clicked.connect(self.on_import)

        self.extra_button_hlayout.insertWidget(0, self.import_button)

    def on_import(self) -> None:
        """Imports all database entries from the database to import into the current database"""
        # creates dicts for every table with the hash of the item as key and the item itself as value
        clips = {hash(clip):clip for clip in self.main_window.database.Clip.get_all()}
        items = {hash(item):item for item in self.main_window.database.Item.get_all()}
        sources = {hash(source):source for source in self.main_window.database.Source.get_all()}
        tags = {hash(tag):tag for tag in self.main_window.database.Tag.get_all()}

        for old_table, new_table in {self.old_database.Clip.get_all(): clips,
                                     self.old_database.Item.get_all(): items,
                                     self.old_database.Source.get_all(): sources,
                                     self.old_database.Tag.get_all(): tags}.items():
            for item in old_table:
                # every item is hashed and checked if it (the hash) already exists
                item_hash = hash(item)
                if item_hash in new_table:
                    # if the item already exists, it's will be deleted
                    logger.warning(f'Skipping {item.table_name} with id {item.id} - already exists in the database to be imported')
                    del new_table[item_hash]

        if clips:
            query = self.old_database.Clip.new_query()
            query.exec(f'INSERT INTO {self.old_database.Clip.table_name} (source_id, path, type, name, description) VALUES '
                       f'({"), (".join([", ".join(self.old_database.Item.make_sql_valid([clip.source_id, clip.path, clip.type.value, clip.name, clip.description])) for clip in clips.values()])})')
            self.old_database.Clip.validate_query(query)
        if items:
            query = self.old_database.Item.new_query()
            query.exec(f'INSERT INTO {self.old_database.Item.table_name} (clip_id, tag_id) VALUES '
                       f'({"), (".join([", ".join(self.old_database.Item.make_sql_valid([item.clip_id, item.tag_id])) for item in items.values()])})')
            self.old_database.Item.validate_query(query)
        if sources:
            query = self.old_database.Source.new_query()
            query.exec(f'INSERT INTO {self.old_database.Source.table_name} (path, type, name, season, episode, description) VALUES '
                       f'({"), (".join([", ".join(self.old_database.Source.make_sql_valid([source.path, source.type.value, source.name, source.season, source.episode, source.description])) for source in sources.values()])})')
            self.old_database.Source.validate_query(query)
        if tags:
            query = self.old_database.Tag.new_query()
            query.exec(f'INSERT INTO {self.old_database.Tag.table_name} (name) VALUES '
                       f'({"), (".join([", ".join(self.old_database.Tag.make_sql_valid([tag.name])) for tag in tags.values()])})')
            self.old_database.Tag.validate_query(query)

        logger.debug(f'Imported {len(clips)} clips, {len(items)} items, {len(sources)} sources, {len(tags)} tags to the database')

        self.reset()
