from . import _DatabaseWindow
from .. import logger
from ..database import Type
from .add import AddSource
from typing import Union
from PyQt5 import QtCore, QtGui, QtWidgets, QtSql


class EditSource(AddSource):
    """Manipulates the `add.AddSource` class to create a edit source class"""

    def __init__(self, window):
        super().__init__(window)

        self.source = None

    def after_setup_ui(self) -> None:
        super().after_setup_ui()

        self.new_button.hide()

    def on_save(self) -> None:
        affected_clips = len(self.source.get_all_clips())
        # if one clip or more is affected
        if affected_clips > 0:
            message = QtWidgets.QMessageBox()
            message.setWindowTitle('Save?')
            message.setText(f'Are you sure to save the edited source? This will affect {affected_clips} {"clips" if affected_clips > 1 else "clip"}')
            affected = message.addButton('Show affected clips', QtWidgets.QMessageBox.ActionRole)
            message.addButton(QtWidgets.QMessageBox.Yes)
            message.addButton(QtWidgets.QMessageBox.No)

            # shows all affected clips
            def on_affected() -> None:
                ShowSource(QtWidgets.QDialog(self.window, QtCore.Qt.WindowSystemMenuHint)).show(self.source.id)

            affected.disconnect()
            affected.clicked.connect(on_affected)

            # shows the message
            result = message.exec()
            if result == QtWidgets.QMessageBox.No:
                return
            elif result == QtWidgets.QMessageBox.Yes:
                type = Type.Source.from_name(self.type_box.currentText()).value
                # edits the source
                self.database.Source.edit(self.source.id, path=self.file_or_url_edit.text(), type=type, name=self.name_edit.text(),
                                          season=self.season_edit.text(), episode=self.episode_edit.text(), description=self.description_edit.toPlainText())

                logger.debug(f'Edited source {self.source.id} - path: `{self.file_or_url_edit.text()}`, type: `{type}`, '
                             f'season: `{self.season_edit.text()}`, episode: {self.episode_edit.text()}, description: {self.description_edit.toPlainText()}')

        self.window.close()

    def show(self, id: int) -> None:
        """
        Shows the source edit

        Args:
            id: The source id to show the source edit
        """
        # gets the source from the id
        self.source = self.database.Source.get(id)

        # sets the texts
        self.file_or_url_edit.setText(self.source.path)
        if self.source.type == Type.Source.AUDIO:
            self.type_box.setCurrentIndex(0)
        elif self.source.type == Type.Source.MOVIE:
            self.type_box.setCurrentIndex(1)
        elif self.source.type == Type.Source.SERIES:
            self.type_box.setCurrentIndex(2)
        else:
            self.type_box.setCurrentIndex(3)
        self.name_edit.setText(self.source.name)
        self.season_edit.setText(self.source.season)
        self.episode_edit.setText(self.source.episode)
        self.description_edit.setText(self.source.description)
        super().show()


class ShowSource(_DatabaseWindow):
    """Class to show all sources"""

    def __init__(self, window):
        self._all_ids = []
        self._source_id = None

        super().__init__(window)

    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(814, 447)
        self.horizontalLayout = QtWidgets.QHBoxLayout(Form)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.main_splitter = QtWidgets.QSplitter(Form)
        self.main_splitter.setOrientation(QtCore.Qt.Horizontal)
        self.main_splitter.setObjectName("main_splitter")
        self.source_frame = QtWidgets.QFrame(self.main_splitter)
        self.source_frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.source_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.source_frame.setObjectName("source_frame")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.source_frame)
        self.verticalLayout.setObjectName("verticalLayout")
        self.source_label = QtWidgets.QLabel(self.source_frame)
        self.source_label.setAlignment(QtCore.Qt.AlignCenter)
        self.source_label.setObjectName("source_label")
        self.verticalLayout.addWidget(self.source_label)
        self.search_edit = QtWidgets.QLineEdit(self.source_frame)
        self.search_edit.setObjectName("search_edit")
        self.verticalLayout.addWidget(self.search_edit)
        self.source_table_view = QtWidgets.QTableView(self.source_frame)
        self.source_table_view.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.source_table_view.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.source_table_view.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.source_table_view.setObjectName("source_table_view")
        self.verticalLayout.addWidget(self.source_table_view)
        self.select_source_button = QtWidgets.QPushButton(self.source_frame)
        self.select_source_button.setAutoDefault(False)
        self.select_source_button.setDefault(False)
        self.select_source_button.setFlat(False)
        self.select_source_button.setObjectName("select_source_button")
        self.verticalLayout.addWidget(self.select_source_button)
        self.all_tags_frame = QtWidgets.QFrame(self.main_splitter)
        self.all_tags_frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.all_tags_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.all_tags_frame.setObjectName("all_tags_frame")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.all_tags_frame)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.all_tags_label = QtWidgets.QLabel(self.all_tags_frame)
        self.all_tags_label.setFrameShadow(QtWidgets.QFrame.Raised)
        self.all_tags_label.setAlignment(QtCore.Qt.AlignCenter)
        self.all_tags_label.setObjectName("all_tags_label")
        self.verticalLayout_2.addWidget(self.all_tags_label)
        self.all_tags_table_view = QtWidgets.QTableView(self.all_tags_frame)
        self.all_tags_table_view.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.all_tags_table_view.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.all_tags_table_view.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.all_tags_table_view.setObjectName("all_tags_table_view")
        self.verticalLayout_2.addWidget(self.all_tags_table_view)
        self.horizontalLayout.addWidget(self.main_splitter)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.source_label.setText(_translate("Form", "Source"))
        self.search_edit.setPlaceholderText(_translate("Form", "Search for source..."))
        self.select_source_button.setText(_translate("Form", "Select source"))
        self.all_tags_label.setText(_translate("Form", "All tags, using this source"))

    def show(self, id: int = None) -> None:
        """
        Shows the window

        Args:
            id: If given, only the source with this is will be shown
        """
        self.set_query(self.database.Source.get(id).query)

        # hides the 'Select' button
        self.select_source_button.hide()

        if id:
            # hides all stuff, which can be used to see other sources than the given
            self.search_edit.hide()

            # select the entry
            index = self.source_table_view.model().index(0, 0)
            self.source_table_view.setCurrentIndex(index)
            self.on_item_clicked(index)
        super().show()

    def set_query(self, query: QtSql.QSqlQuery) -> None:
        """
        Sets the source query

        Args:
            query: The source query
        """
        # the sql model
        model = QtSql.QSqlQueryModel()
        model.setQuery(query)

        # sets the model to the source table view
        self.source_table_view.setModel(model)

        for i in range(model.columnCount()):
            header = str(model.headerData(i, QtCore.Qt.Horizontal)).lower()
            # hides columns with these headers
            if header in ['id', 'name', 'season', 'episode']:
                if header == 'id':
                    for row in range(model.rowCount()):
                        self._all_ids.append(int(model.index(row, i).data()))
                self.source_table_view.hideColumn(i)
            else:
                model.setHeaderData(i, QtCore.Qt.Horizontal, header[0].upper() + header[1:])

        # connect all button click events
        self.select_source_button.clicked.connect(self.on_select_source)
        self.source_table_view.clicked.connect(self.on_item_clicked)
        self.source_table_view.doubleClicked.connect(self.on_select_source)

    def on_item_clicked(self, index: QtCore.QModelIndex):
        """
        Gets called if a source item is clicked

        Args:
            index: The item which was clicked
        """

        row = index.row()
        # the source id of the item
        self._source_id = self._all_ids[row - 1]

        # the model for all tags which are using the selected source
        model = QtSql.QSqlTableModel()
        model.setQuery(self.database.Clip.get_all_by(source_id=self._source_id).query)
        model.setEditStrategy(QtSql.QSqlTableModel.OnManualSubmit)

        # sets the model
        self.all_tags_table_view.setModel(model)
        self.all_tags_table_view.horizontalHeader().resizeSections(QtWidgets.QHeaderView.Stretch)
        self.all_tags_table_view.horizontalHeader().setStretchLastSection(True)

        # this sets the header (first letter uppercase, remove id, etc.)
        for i in range(model.columnCount()):
            header = str(model.headerData(i, QtCore.Qt.Horizontal)).lower()
            # » hide the header with the names id or source_id
            if header in ['id', 'source_id']:
                if header == 'id':
                    for row in range(model.rowCount()):
                        self._all_ids.append(int(model.index(row, i).data()))
                self.all_tags_table_view.hideColumn(i)
                continue
            elif header == 'type':
                for row in range(model.rowCount()):
                    index = model.index(row, i)
                    type_name = Type.Clip(model.data(index)).name.lower()
                    model.setData(index, type_name[0].upper() + type_name[1:])
                    model.submit()
                    model.setData(index, QtGui.QColor(100, 100, 100, 1), QtCore.Qt.BackgroundColorRole)
                    model.submit()
            # » first letter of the header uppercase
            model.setHeaderData(i, QtCore.Qt.Horizontal, header[0].upper() + header[1:])

    def on_select_source(self):
        """Gets called if the 'Select source' button is pressed"""
        if not self._source_id:
            proceed = QtWidgets.QMessageBox.question(self.window, 'No source is chosen', 'No source is chosen. Do you want to proceed?',
                                                     QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
            if proceed == QtWidgets.QMessageBox.No:
                return
        self.window.close()

    def select(self) -> Union[int, None]:
        """
        Shows the window and enable the user to select a source

        Returns:
            The selected source id
        """
        self.set_query(self.database.Source.get_all().query)
        self.window.exec()
        return self._source_id
